from .types import (
    OpenBESSpecification,
    OpenBESParameters,
    ENERGY_SOURCES,
    COOLING_SYSTEM_TYPES,
    HEAT_CAPACITY_CLASSES,
    HEATING_SYSTEM_TYPES,
    LIGHTING_CONTROL,
    LIGHTING_BALLASTS,
    LIGHTING_TECHNOLOGIES,
    TERRAINS
)

HOLYWELL_HOUSE_SPEC = OpenBESSpecification(
    parameters=OpenBESParameters(
        temperature_tolerance=0
    ),
    appliances_load=1.0,
    biomass_annual=None,
    biomass_pellets_annual=None,
    building_area=None,
    building_height=5.5,
    building_length=43.41,
    building_name=None,
    building_standby_load=2321.2,
    building_type=None,
    building_width=14.13,
    condition_z1=True,
    condition_z2=True,
    condition_z3=True,
    condition_z4=True,
    condition_z5=False,
    cooling_system1_energy_efficifiency_ratio=2.8,
    cooling_system1_energy_source=ENERGY_SOURCES.Electricity,
    cooling_system1_nominal_capacity=96.0,
    cooling_system1_number=1,
    cooling_system1_off_time=17,
    cooling_system1_on_time=8,
    cooling_system1_sensible_nominal_capacity=75.0,
    cooling_system1_simultaneity_factor_canteen=1.0,
    cooling_system1_simultaneity_factor_common=0.0,
    cooling_system1_simultaneity_factor_office=1.0,
    cooling_system1_simultaneity_factor_other=0.0,
    cooling_system1_simultaneity_factor_teaching=1.0,
    cooling_system1_type=COOLING_SYSTEM_TYPES.Heat_pump,
    country=None,
    diesel_annual=None,
    electricity_annual=None,
    electricity_april=None,
    electricity_august=None,
    electricity_december=None,
    electricity_february=None,
    electricity_january=None,
    electricity_july=None,
    electricity_june=None,
    electricity_march=None,
    electricity_may=None,
    electricity_november=None,
    electricity_october=None,
    electricity_september=None,
    energy_generated=None,
    energy_used=None,
    first_floor_area_z1=494.85,
    first_floor_area_z2=None,
    first_floor_area_z3=None,
    first_floor_area_z4=55.65,
    first_floor_area_z5=None,
    floor_to_ceiling_height=2.90,
    fourth_floor_area_z1=None,
    fourth_floor_area_z2=None,
    fourth_floor_area_z3=None,
    fourth_floor_area_z4=None,
    fourth_floor_area_z5=None,
    gas_april=None,
    gas_august=None,
    gas_december=None,
    gas_february=None,
    gas_january=None,
    gas_july=None,
    gas_june=None,
    gas_march=None,
    gas_may=None,
    gas_november=None,
    gas_october=None,
    gas_september=None,
    ground_floor_area_z1=190.56,
    ground_floor_area_z2=129.28,
    ground_floor_area_z3=97.88,
    ground_floor_area_z4=132.76,
    ground_floor_area_z5=None,
    heat_capacity=HEAT_CAPACITY_CLASSES.Medium,
    heating_system1_efficiency_cop=0.86,
    heating_system1_energy_source=ENERGY_SOURCES.Natural_gas,
    heating_system1_nominal_capacity=48.0,
    heating_system1_number=2,
    heating_system1_off_time=None,
    heating_system1_on_time=None,
    heating_system1_simultaneity_factor_canteen=1.0,
    heating_system1_simultaneity_factor_common=0.2,
    heating_system1_simultaneity_factor_office=1.0,
    heating_system1_simultaneity_factor_other=0.0,
    heating_system1_simultaneity_factor_teaching=1.0,
    heating_system1_type=HEATING_SYSTEM_TYPES.Condensing_boiler,
    holiday=False,
    leakage_air_flow=None,
    leakage_air_flow_independent=0.45,
    lighting_control=LIGHTING_CONTROL.Manual,
    lighting_off_time=17,
    lighting_on_time=9,
    lighting_simultaneity_factor=0.2,
    lighting_system_ballast_z1=LIGHTING_BALLASTS.BE,
    lighting_system_ballast_z2=None,
    lighting_system_ballast_z3=None,
    lighting_system_ballast_z4=None,
    lighting_system_ballast_z5=None,
    lighting_system_ballast_z6=None,
    lighting_system_lamp_number_z1=4,
    lighting_system_lamp_number_z2=1,
    lighting_system_lamp_number_z3=None,
    lighting_system_lamp_number_z4=None,
    lighting_system_lamp_number_z5=None,
    lighting_system_lamp_number_z6=None,
    lighting_system_lamp_power_z1=18,
    lighting_system_lamp_power_z2=40,
    lighting_system_lamp_power_z3=None,
    lighting_system_lamp_power_z4=None,
    lighting_system_lamp_power_z5=None,
    lighting_system_lamp_power_z6=None,
    lighting_system_luminary_number_z1=35,
    lighting_system_luminary_number_z2=55,
    lighting_system_luminary_number_z3=None,
    lighting_system_luminary_number_z4=None,
    lighting_system_luminary_number_z5=None,
    lighting_system_luminary_number_z6=None,
    lighting_system_name_z1="First Floor",
    lighting_system_name_z2="Second Floor",
    lighting_system_name_z3=None,
    lighting_system_name_z4=None,
    lighting_system_name_z5=None,
    lighting_system_name_z6=None,
    lighting_system_operating_hours_z1=8,
    lighting_system_operating_hours_z2=8,
    lighting_system_operating_hours_z3=None,
    lighting_system_operating_hours_z4=None,
    lighting_system_operating_hours_z5=None,
    lighting_system_operating_hours_z6=None,
    lighting_system_similar_zone_number_z1=1,
    lighting_system_similar_zone_number_z2=1,
    lighting_system_similar_zone_number_z3=None,
    lighting_system_similar_zone_number_z4=None,
    lighting_system_similar_zone_number_z5=None,
    lighting_system_similar_zone_number_z6=None,
    lighting_system_simultaneity_factor_z1=0.7,
    lighting_system_simultaneity_factor_z2=0.7,
    lighting_system_simultaneity_factor_z3=None,
    lighting_system_simultaneity_factor_z4=None,
    lighting_system_simultaneity_factor_z5=None,
    lighting_system_simultaneity_factor_z6=None,
    lighting_system_tech_z1=LIGHTING_TECHNOLOGIES.FT_T8,
    lighting_system_tech_z2=LIGHTING_TECHNOLOGIES.LED,
    lighting_system_tech_z3=None,
    lighting_system_tech_z4=None,
    lighting_system_tech_z5=None,
    lighting_system_tech_z6=None,
    location=None,
    LPG_annual=None,
    max_building_occupation=150,
    meteorological_file="UK_Oxford_GBR_ENG_RAF.Benson.036580_TMYx.2007-2021.epw",
    natural_gas_annual=None,
    natural_ventilation_night=0.0,
    occupancy_close_canteen=14,
    occupancy_close_office=17,
    occupancy_close_teaching=17,
    occupancy_open_canteen=10,
    occupancy_open_office=8,
    occupancy_open_teaching=9,
    orientation_angle=165.0,
    other_electricity_usage=1136.0,
    other_gas_usage=0.0,
    roof_angle=35,
    schedule_april=True,
    schedule_august=True,
    schedule_december=True,
    schedule_february=True,
    schedule_friday=True,
    schedule_january=True,
    schedule_july=True,
    schedule_june=True,
    schedule_march=True,
    schedule_may=True,
    schedule_monday=True,
    schedule_november=True,
    schedule_october=True,
    schedule_saturday=False,
    schedule_september=True,
    schedule_sunday=False,
    schedule_thursday=True,
    schedule_tuesday=True,
    schedule_wednesday=True,
    second_floor_area_z1=None,
    second_floor_area_z2=None,
    second_floor_area_z3=None,
    second_floor_area_z4=52.93,
    second_floor_area_z5=None,
    setpoint_summer_day=21,
    setpoint_summer_night=29,
    setpoint_winter_day=22,
    setpoint_winter_night=18,
    slab_thickness=0.35,
    solar_external_shading_summer=0.9,
    solar_external_shading_winter=0.6,
    terrain_class=TERRAINS.Country,
    thermal_bridge_facade_ground=True,
    thermal_bridge_facade_intermediate=False,
    thermal_bridge_facade_roof=True,
    thermal_bridge_shading=False,
    thermal_bridge_window=False,
    third_floor_area_z1=None,
    third_floor_area_z2=None,
    third_floor_area_z3=None,
    third_floor_area_z4=None,
    third_floor_area_z5=None,
    typical_occupation=75,
    uvalue_facade=0.97,
    uvalue_floor=1.368,
    uvalue_roof=0.84,
    uvalue_window=2.5,
    ventilation_system1_airflow=300.0,
    ventilation_system1_energy_source=ENERGY_SOURCES.Electricity,
    ventilation_system1_heat_recovery_efficiency=0.75,
    ventilation_system1_off_time=14,
    ventilation_system1_on_time=10,
    ventilation_system1_rated_input_power=0.3,
    ventilation_system1_type=None,
    ventilation_system1_ventilated_area=100.0,
    water_demand=300.0,
    water_reference_temperature=60.0,
    water_supply_temperature=16.0,
    water_system_efficiency_cop=1.0,
    water_system_energy_source=ENERGY_SOURCES.Electricity,
    water_system_nominal_capacity=None,
    water_system_type=None,
    window_frame_factor=0.15,
    window_gvalue=0.65,
    window_height=1.6,
    window_length=2.4,
    window_number_first_a1=9,
    window_number_first_b1=3,
    window_number_first_c1=9,
    window_number_first_d1=None,
    window_number_fourth_a1=None,
    window_number_fourth_b1=None,
    window_number_fourth_c1=None,
    window_number_fourth_d1=None,
    window_number_ground_a1=9,
    window_number_ground_b1=3,
    window_number_ground_c1=9,
    window_number_ground_d1=None,
    window_number_second_a1=None,
    window_number_second_b1=None,
    window_number_second_c1=None,
    window_number_second_d1=None,
    window_number_third_a1=None,
    window_number_third_b1=None,
    window_number_third_c1=None,
    window_number_third_d1=None,
    zone_name_z1=None,
    zone_name_z2=None,
    zone_name_z3=None,
    zone_name_z4=None,
    zone_name_z5=None,
)
